class ClientConfig(object):
    PUBLIC_KEY = 'VllVpxXLMCpE5Uuw1nqkLDPjGE85Pnmpdt8VIcmOr40'
    APP_NAME = 'kubectl-val'
    COMPANY_NAME = 'CriticalHop'
    HTTP_TIMEOUT = 6000
    MAX_DOWNLOAD_RETRIES = 3
    UPDATE_URLS = ['https://download.criticalhop.com/']
